#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

META_EDGE = "edge"
META_ITERATION = "iteration"
META_BRACKET_ITERATION = "bracket_iteration"
META_REWRITE_PATH = "rewrite_path"
META_HAS_EARLY_STOPPING = "has_early_stopping"
META_HAS_JOBS = "has_jobs"
META_HAS_SERVICES = "has_services"
META_HAS_DAGS = "has_dags"
META_HAS_MATRICES = "has_matrices"
META_HAS_HOOKS = "has_hooks"
META_IS_HOOK = "is_hook"
META_IS_ACTION = "is_action"
META_IS_TUNER = "is_tuner"
META_CONCURRENCY = "concurrency"
META_HAS_DOWNSTREAM_EVENTS_TRIGGER = "has_downstream_events_trigger"
